@@ -7,6 +7,7 @@
 // except according to those terms.
 
 use std::cmp;
+use std::ptr as std_ptr;
 use std::slice;
 
 use imp_prelude::*;
@@ -29,6 +30,7 @@ use {
     AxisIter,
     AxisIterMut,
 };
+use stacking::stack;
 
 impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
 {
@@ -89,7 +91,7 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
             (self.iter().cloned().collect(), self.dim.default_strides())
         };
         unsafe {
-            ArrayBase::from_vec_dim_stride_unchecked(self.dim.clone(), strides, data)
+            ArrayBase::from_shape_vec_unchecked(self.dim.clone().strides(strides), data)
         }
     }
 
@@ -196,6 +198,8 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
         debug_assert!(self.pointer_is_inbounds());
     }
 
+
+
     /// Return a reference to the element at `index`, or return `None`
     /// if the index is out of bounds.
     ///
@@ -261,6 +265,22 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
         &mut *self.ptr.offset(off)
     }
 
+    /// Swap elements at indices `index1` and `index2`.
+    ///
+    /// Indices may be equal.
+    ///
+    /// ***Panics*** if an index is out of bounds.
+    pub fn swap<I>(&mut self, index1: I, index2: I)
+        where S: DataMut,
+              I: NdIndex<Dim=D>,
+    {
+        let ptr1: *mut _ = &mut self[index1];
+        let ptr2: *mut _ = &mut self[index2];
+        unsafe {
+            std_ptr::swap(ptr1, ptr2);
+        }
+    }
+
     // `get` for zero-dimensional arrays
     // panics if dimension is not zero. otherwise an element is always present.
     fn get_0d(&self) -> &A {
@@ -328,7 +348,10 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
     /// let mut a = arr2(&[[1., 2.],
     ///                    [3., 4.]]);
     ///
-    /// a.subview_mut(Axis(1), 1).iadd_scalar(&10.);
+    /// {
+    ///     let mut column1 = a.subview_mut(Axis(1), 1);
+    ///     column1 += 10.;
+    /// }
     ///
     /// assert!(
     ///     a == aview2(&[[1., 12.],
@@ -371,6 +394,46 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
         }
     }
 
+    /// Along `axis`, select arbitrary subviews corresponding to `indices`
+    /// and and copy them into a new array.
+    ///
+    /// **Panics** if `axis` or an element of `indices` is out of bounds.
+    ///
+    /// ```
+    /// use ndarray::{arr2, Axis};
+    ///
+    /// let x = arr2(&[[0., 1.],
+    ///                [2., 3.],
+    ///                [4., 5.],
+    ///                [6., 7.],
+    ///                [8., 9.]]);
+    ///
+    /// let r = x.select(Axis(0), &[0, 4, 3]);
+    /// assert!(
+    ///         r == arr2(&[[0., 1.],
+    ///                     [8., 9.],
+    ///                     [6., 7.]])
+    ///);
+    /// ```
+    pub fn select(&self, axis: Axis, indices: &[Ix]) -> OwnedArray<A, D>
+        where A: Copy,
+              D: RemoveAxis,
+    {
+        let mut subs = vec![self.view(); indices.len()];
+        for (&i, sub) in zipsl(indices, &mut subs[..]) {
+            sub.isubview(axis, i);
+        }
+        if subs.is_empty() {
+            let mut dim = self.dim();
+            dim.set_axis(axis, 0);
+            unsafe {
+                OwnedArray::from_shape_vec_unchecked(dim, vec![])
+            }
+        } else {
+            stack(axis, &subs).unwrap()
+        }
+    }
+
     /// Return an iterator that traverses over all dimensions but the innermost,
     /// and yields each inner row.
     ///
@@ -616,7 +679,7 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
     /// Return a pointer to the first element in the array.
     ///
     /// Raw access to array elements needs to follow the strided indexing
-    /// scheme: an element at multi-index *I* in an array with strides *S* is 
+    /// scheme: an element at multi-index *I* in an array with strides *S* is
     /// located at offset
     ///
     /// *Σ<sub>0 ≤ k < d</sub> I<sub>k</sub> × S<sub>k</sub>*
@@ -737,7 +800,7 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
         } else {
             let v = self.iter().map(|x| x.clone()).collect::<Vec<A>>();
             unsafe {
-                ArrayBase::from_vec_dim_unchecked(shape, v)
+                ArrayBase::from_shape_vec_unchecked(shape, v)
             }
         }
     }
@@ -747,8 +810,7 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
     /// contiguous, otherwise we cannot rearrange the dimension.
     ///
     /// **Errors** if the shapes don't have the same number of elements.<br>
-    /// **Errors** if the input array is not c-contiguous (this will be
-    /// slightly improved in the future).
+    /// **Errors** if the input array is not c- or f-contiguous.
     ///
     /// ```
     /// use ndarray::{aview1, aview2};
@@ -773,6 +835,13 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
                 strides: shape.default_strides(),
                 dim: shape,
             })
+        } else if self.ndim() > 1 && self.view().reversed_axes().is_standard_layout() {
+            Ok(ArrayBase {
+                data: self.data,
+                ptr: self.ptr,
+                strides: shape.fortran_strides(),
+                dim: shape,
+            })
         } else {
             Err(error::from_kind(error::ErrorKind::IncompatibleLayout))
         }
@@ -903,40 +972,8 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
         self.view().reversed_axes()
     }
 
-    /// ***Deprecated: Use .as_slice_memory_order() instead.***
-    ///
-    /// Return a slice of the array’s backing data in memory order.
-    ///
-    /// **Note:** Data memory order may not correspond to the index order
-    /// of the array. Neither is the raw data slice is restricted to just the
-    /// array’s view.<br>
-    #[cfg_attr(has_deprecated, deprecated(note="Use .as_slice_memory_order() instead"))]
-    pub fn raw_data(&self) -> &[A]
-        where S: DataOwned,
-    {
-        self.data.slice()
-    }
-
-    /// ***Deprecated: Use .as_slice_memory_order_mut() instead.***
-    ///
-    /// Return a mutable slice of the array’s backing data in memory order.
-    ///
-    /// **Note:** Data memory order may not correspond to the index order
-    /// of the array. Neither is the raw data slice is restricted to just the
-    /// array’s view.<br>
-    ///
-    /// **Note:** The data is uniquely held and nonaliased
-    /// while it is mutably borrowed.
-    #[cfg_attr(has_deprecated, deprecated(note="Use .as_slice_memory_order_mut() instead"))]
-    pub fn raw_data_mut(&mut self) -> &mut [A]
-        where S: DataOwned + DataMut,
-    {
-        self.ensure_unique();
-        self.data.slice_mut()
-    }
-
     fn pointer_is_inbounds(&self) -> bool {
-        let slc = self.data.slice();
+        let slc = self.data._data_slice();
         if slc.is_empty() {
             // special case for data-less views
             return true;
@@ -1066,16 +1103,14 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
         }
     }
 
-    /// ***Deprecated: Will be removed because it dictates a specific order.***
-    ///
-    /// Traverse the array elements in order and apply a fold,
+    /// Traverse the array elements and apply a fold,
     /// returning the resulting value.
-    #[cfg_attr(has_deprecated, deprecated(note=
-      "Will be removed because it dictates a specific order"))]
+    ///
+    /// Elements are visited in arbitrary order.
     pub fn fold<'a, F, B>(&'a self, mut init: B, mut f: F) -> B
         where F: FnMut(B, &'a A) -> B, A: 'a
     {
-        if let Some(slc) = self.as_slice() {
+        if let Some(slc) = self.as_slice_memory_order() {
             // FIXME: Use for loop when slice iterator is perf is restored
             for i in 0..slc.len() {
                 init = f(init, &slc[i]);
@@ -1115,13 +1150,13 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
         if let Some(slc) = self.as_slice_memory_order() {
             let v = ::iterators::to_vec(slc.iter().map(f));
             unsafe {
-                ArrayBase::from_vec_dim_stride_unchecked(
-                    self.dim.clone(), self.strides.clone(), v)
+                ArrayBase::from_shape_vec_unchecked(
+                    self.dim.clone().strides(self.strides.clone()), v)
             }
         } else {
             let v = ::iterators::to_vec(self.iter().map(f));
             unsafe {
-                ArrayBase::from_vec_dim_unchecked(self.dim.clone(), v)
+                ArrayBase::from_shape_vec_unchecked(self.dim.clone(), v)
             }
         }
     }
@@ -1224,4 +1259,23 @@ impl<A, S, D> ArrayBase<S, D> where S: Data<Elem=A>, D: Dimension
             }
         }
     }
+
+    /// Fold along an axis.
+    ///
+    /// Combine the elements of each subview with the previous using the `fold`
+    /// function and initial value `init`.
+    ///
+    /// Return the result as an `OwnedArray`.
+    pub fn fold_axis<B, F>(&self, axis: Axis, init: B, mut fold: F)
+        -> OwnedArray<B, D::Smaller>
+        where D: RemoveAxis,
+              F: FnMut(&B, &A) -> B,
+              B: Clone,
+    {
+        let mut res = OwnedArray::from_elem(self.dim().remove_axis(axis), init);
+        for subview in self.axis_iter(axis) {
+            res.zip_mut_with(&subview, |x, y| *x = fold(x, y));
+        }
+        res
+    }
 }
