@@ -1,15 +1,10 @@
 #[macro_use(s)] extern crate ndarray;
 extern crate num as libnum;
 
-use ndarray::RcArray;
+use ndarray::prelude::*;
 use ndarray::{arr0, rcarr1, rcarr2};
-use ndarray::{
-    OwnedArray,
-    Ix, Ixs,
-    AsArray,
-    NdFloat,
-    aview0,
-};
+use ndarray::{LinalgScalar, Data};
+use ndarray::linalg::general_mat_mul;
 
 use std::fmt;
 use libnum::Float;
@@ -32,39 +27,37 @@ fn test_oper(op: &str, a: &[f32], b: &[f32], c: &[f32])
     test_oper_arr(op, aa.clone(), bb.clone(), cc.clone());
 }
 
-fn test_oper_arr<A: Float + fmt::Debug, D: ndarray::Dimension>
+fn test_oper_arr<A: NdFloat + fmt::Debug, D: ndarray::Dimension>
     (op: &str, mut aa: RcArray<A,D>, bb: RcArray<A, D>, cc: RcArray<A, D>)
 {
     match op {
         "+" => {
             assert_eq!(&aa + &bb, cc);
-            aa.iadd(&bb);
+            aa += &bb;
             assert_eq!(aa, cc);
         },
         "-" => {
             assert_eq!(&aa - &bb, cc);
-            aa.isub(&bb);
+            aa -= &bb;
             assert_eq!(aa, cc);
         },
         "*" => {
             assert_eq!(&aa * &bb, cc);
-            aa.imul(&bb);
+            aa *= &bb;
             assert_eq!(aa, cc);
         },
         "/" => {
             assert_eq!(&aa / &bb, cc);
-            aa.idiv(&bb);
+            aa /= &bb;
             assert_eq!(aa, cc);
         },
         "%" => {
             assert_eq!(&aa % &bb, cc);
-            aa.irem(&bb);
+            aa %= &bb;
             assert_eq!(aa, cc);
         },
         "neg" => {
             assert_eq!(-aa.clone(), cc);
-            aa.ineg();
-            assert_eq!(aa, cc);
         },
         _ => panic!()
     }
@@ -91,7 +84,7 @@ fn scalar_operations()
     {
         let mut x = a.clone();
         let mut y = arr0(0.);
-        x.iadd_scalar(&1.);
+        x += 1.;
         y.assign_scalar(&2.);
         assert_eq!(x, a + arr0(1.));
         assert_eq!(x, y);
@@ -100,7 +93,7 @@ fn scalar_operations()
     {
         let mut x = b.clone();
         let mut y = rcarr1(&[0., 0.]);
-        x.iadd_scalar(&1.);
+        x += 1.;
         y.assign_scalar(&2.);
         assert_eq!(x, b + arr0(1.));
         assert_eq!(x, y);
@@ -109,7 +102,7 @@ fn scalar_operations()
     {
         let mut x = c.clone();
         let mut y = RcArray::zeros((2, 2));
-        x.iadd_scalar(&1.);
+        x += 1.;
         y.assign_scalar(&2.);
         assert_eq!(x, c + arr0(1.));
         assert_eq!(x, y);
@@ -122,6 +115,25 @@ fn assert_approx_eq<F: fmt::Debug + Float>(f: F, g: F, tol: F) -> bool {
     true
 }
 
+fn assert_close<D>(a: ArrayView<f64, D>, b: ArrayView<f64, D>)
+    where D: Dimension,
+{
+    let diff = (&a - &b).mapv_into(f64::abs);
+
+    let rtol = 1e-7;
+    let atol = 1e-12;
+    let crtol = b.mapv(|x| x.abs() * rtol);
+    let tol = crtol + atol;
+    let tol_m_diff = &diff - &tol;
+    let maxdiff = tol_m_diff.fold(0./0., |x, y| f64::max(x, *y));
+    println!("diff offset from tolerance level= {:.2e}", maxdiff);
+    if maxdiff > 0. {
+        println!("{:.4?}", a);
+        println!("{:.4?}", b);
+        panic!("results differ");
+    }
+}
+
 fn reference_dot<'a,A, V1, V2>(a: V1, b: V2) -> A
     where A: NdFloat,
           V1: AsArray<'a, A>,
@@ -216,7 +228,47 @@ fn range_mat(m: Ix, n: Ix) -> OwnedArray<f32, (Ix, Ix)> {
     OwnedArray::linspace(0., (m * n - 1) as f32, m * n).into_shape((m, n)).unwrap()
 }
 
-#[cfg(has_assign)]
+fn range_mat64(m: Ix, n: Ix) -> OwnedArray<f64, (Ix, Ix)> {
+    OwnedArray::linspace(0., (m * n - 1) as f64, m * n).into_shape((m, n)).unwrap()
+}
+
+fn range_i32(m: Ix, n: Ix) -> OwnedArray<i32, (Ix, Ix)> {
+    OwnedArray::from_iter(0..(m * n) as i32).into_shape((m, n)).unwrap()
+}
+
+// simple, slow, correct (hopefully) mat mul
+fn reference_mat_mul<A, S, S2>(lhs: &ArrayBase<S, (Ix, Ix)>, rhs: &ArrayBase<S2, (Ix, Ix)>)
+    -> OwnedArray<A, (Ix, Ix)>
+    where A: LinalgScalar,
+          S: Data<Elem=A>,
+          S2: Data<Elem=A>,
+{
+    let ((m, k), (k2, n)) = (lhs.dim(), rhs.dim());
+    assert!(m.checked_mul(n).is_some());
+    assert_eq!(k, k2);
+    let mut res_elems = Vec::<A>::with_capacity(m * n);
+    unsafe {
+        res_elems.set_len(m * n);
+    }
+
+    let mut i = 0;
+    let mut j = 0;
+    for rr in &mut res_elems {
+        unsafe {
+            *rr = (0..k).fold(A::zero(),
+                move |s, x| s + *lhs.uget((i, x)) * *rhs.uget((x, j)));
+        }
+        j += 1;
+        if j == n {
+            j = 0;
+            i += 1;
+        }
+    }
+    unsafe {
+        ArrayBase::from_shape_vec_unchecked((m, n), res_elems)
+    }
+}
+
 #[test]
 fn mat_mul() {
     let (m, n, k) = (8, 8, 8);
@@ -227,16 +279,16 @@ fn mat_mul() {
         let mut c = b.column_mut(0);
         c += 1.0;
     }
-    let ab = a.mat_mul(&b);
+    let ab = a.dot(&b);
 
-    let mut af = OwnedArray::zeros_f(a.dim());
-    let mut bf = OwnedArray::zeros_f(b.dim());
+    let mut af = OwnedArray::zeros(a.dim().f());
+    let mut bf = OwnedArray::zeros(b.dim().f());
     af.assign(&a);
     bf.assign(&b);
 
-    assert_eq!(ab, a.mat_mul(&bf));
-    assert_eq!(ab, af.mat_mul(&b));
-    assert_eq!(ab, af.mat_mul(&bf));
+    assert_eq!(ab, a.dot(&bf));
+    assert_eq!(ab, af.dot(&b));
+    assert_eq!(ab, af.dot(&bf));
 
     let (m, n, k) = (10, 5, 11);
     let a = range_mat(m, n);
@@ -246,16 +298,16 @@ fn mat_mul() {
         let mut c = b.column_mut(0);
         c += 1.0;
     }
-    let ab = a.mat_mul(&b);
+    let ab = a.dot(&b);
 
-    let mut af = OwnedArray::zeros_f(a.dim());
-    let mut bf = OwnedArray::zeros_f(b.dim());
+    let mut af = OwnedArray::zeros(a.dim().f());
+    let mut bf = OwnedArray::zeros(b.dim().f());
     af.assign(&a);
     bf.assign(&b);
 
-    assert_eq!(ab, a.mat_mul(&bf));
-    assert_eq!(ab, af.mat_mul(&b));
-    assert_eq!(ab, af.mat_mul(&bf));
+    assert_eq!(ab, a.dot(&bf));
+    assert_eq!(ab, af.dot(&b));
+    assert_eq!(ab, af.dot(&bf));
 
     let (m, n, k) = (10, 8, 1);
     let a = range_mat(m, n);
@@ -265,16 +317,16 @@ fn mat_mul() {
         let mut c = b.column_mut(0);
         c += 1.0;
     }
-    let ab = a.mat_mul(&b);
+    let ab = a.dot(&b);
 
-    let mut af = OwnedArray::zeros_f(a.dim());
-    let mut bf = OwnedArray::zeros_f(b.dim());
+    let mut af = OwnedArray::zeros(a.dim().f());
+    let mut bf = OwnedArray::zeros(b.dim().f());
     af.assign(&a);
     bf.assign(&b);
 
-    assert_eq!(ab, a.mat_mul(&bf));
-    assert_eq!(ab, af.mat_mul(&b));
-    assert_eq!(ab, af.mat_mul(&bf));
+    assert_eq!(ab, a.dot(&bf));
+    assert_eq!(ab, af.dot(&b));
+    assert_eq!(ab, af.dot(&bf));
 }
 
 // Check that matrix multiplication of contiguous matrices returns a
@@ -284,18 +336,39 @@ fn mat_mul_order() {
     let (m, n, k) = (8, 8, 8);
     let a = range_mat(m, n);
     let b = range_mat(n, k);
-    let mut af = OwnedArray::zeros_f(a.dim());
-    let mut bf = OwnedArray::zeros_f(b.dim());
+    let mut af = OwnedArray::zeros(a.dim().f());
+    let mut bf = OwnedArray::zeros(b.dim().f());
     af.assign(&a);
     bf.assign(&b);
 
-    let cc = a.mat_mul(&b);
-    let ff = af.mat_mul(&bf);
+    let cc = a.dot(&b);
+    let ff = af.dot(&bf);
 
     assert_eq!(cc.strides()[1], 1);
     assert_eq!(ff.strides()[0], 1);
 }
 
+// test matrix multiplication shape mismatch
+#[test]
+#[should_panic]
+fn mat_mul_shape_mismatch() {
+    let (m, k, k2, n) = (8, 8, 9, 8);
+    let a = range_mat(m, k);
+    let b = range_mat(k2, n);
+    a.dot(&b);
+}
+
+// test matrix multiplication shape mismatch
+#[test]
+#[should_panic]
+fn mat_mul_shape_mismatch_2() {
+    let (m, k, k2, n) = (8, 8, 8, 8);
+    let a = range_mat(m, k);
+    let b = range_mat(k2, n);
+    let mut c = range_mat(m, n + 1);
+    general_mat_mul(1., &a, &b, 1., &mut c);
+}
+
 // Check that matrix multiplication
 // supports broadcast arrays.
 #[test]
@@ -309,9 +382,9 @@ fn mat_mul_broadcast() {
     let b1 = b1.broadcast((n, k)).unwrap();
     let b2 = OwnedArray::from_elem((n, k), x1);
 
-    let c2 = a.mat_mul(&b2);
-    let c1 = a.mat_mul(&b1);
-    let c0 = a.mat_mul(&b0);
+    let c2 = a.dot(&b2);
+    let c1 = a.dot(&b1);
+    let c0 = a.dot(&b0);
     assert_eq!(c2, c1);
     assert_eq!(c2, c0);
 }
@@ -327,8 +400,8 @@ fn mat_mul_rev() {
     rev.assign(&b);
     println!("{:.?}", rev);
 
-    let c1 = a.mat_mul(&b);
-    let c2 = a.mat_mul(&rev);
+    let c1 = a.dot(&b);
+    let c2 = a.dot(&rev);
     assert_eq!(c1, c2);
 }
 
@@ -346,3 +419,65 @@ fn scaled_add() {
     assert_eq!(c, d);
 
 }
+
+#[test]
+fn gen_mat_mul() {
+    let alpha = -2.3;
+    let beta = 3.14;
+    let sizes = vec![(4, 4, 4), (8, 8, 8),
+                     (17, 15, 16),
+                     (4, 17, 3),
+                     (17, 3, 22),
+                     (19, 18, 2),
+                     (16, 17, 15),
+                     (15, 16, 17),
+                     (67, 63, 62),
+        ];
+    // test different strides
+    for &s1 in &[1, 2, -1, -2] {
+        for &s2 in &[1, 2, -1, -2] {
+            for &(m, k, n) in &sizes {
+                let a = range_mat64(m, k);
+                let b = range_mat64(k, n);
+                let mut c = range_mat64(m, n);
+                let mut answer = c.clone();
+
+                {
+                    let a = a.slice(s![..;s1, ..;s2]);
+                    let b = b.slice(s![..;s2, ..;s2]);
+                    let mut cv = c.slice_mut(s![..;s1, ..;s2]);
+
+                    let answer_part = alpha * reference_mat_mul(&a, &b) + beta * &cv;
+                    answer.slice_mut(s![..;s1, ..;s2]).assign(&answer_part);
+
+                    general_mat_mul(alpha, &a, &b, beta, &mut cv);
+                }
+                assert_close(c.view(), answer.view());
+            }
+        }
+    }
+}
+
+#[test]
+fn gen_mat_mul_i32() {
+    let alpha = -1;
+    let beta = 2;
+    let sizes = vec![(4, 4, 4), (8, 8, 8),
+                     (17, 15, 16),
+                     (4, 17, 3),
+                     (17, 3, 22),
+                     (19, 18, 2),
+                     (16, 17, 15),
+                     (15, 16, 17),
+                     (67, 63, 62),
+        ];
+    for &(m, k, n) in &sizes {
+        let a = range_i32(m, k);
+        let b = range_i32(k, n);
+        let mut c = range_i32(m, n);
+
+        let answer = alpha * reference_mat_mul(&a, &b) + beta * &c;
+        general_mat_mul(alpha, &a, &b, beta, &mut c);
+        assert_eq!(&c, &answer);
+    }
+}
